import json
from copy import copy
from typing import List, Optional

import tiktoken
from openai.types.chat.chat_completion_message import ChatCompletionMessage


class ChatgptEvent:
    SYSTEM_MESSAGE = "system_message"
    ASSISTENT_MESSAGE = "assistent_message"
    USER_MESSAGE = "user_message"
    TOOL_CALL_REQUEST = "tool_call_request"
    TOOL_CALL_RESULT = "tool_call_result"
    list_event_type = [SYSTEM_MESSAGE, ASSISTENT_MESSAGE, USER_MESSAGE, TOOL_CALL_REQUEST, TOOL_CALL_RESULT]

    def __init__(
        self,
        event_type: str,
        event_message: dict,
        *,
        list_tool_offer: Optional[list[dict]] = None,
        tool_choice: Optional[str] = None,
        response_format: Optional[dict] = None,
    ) -> None:
        if event_type not in ChatgptEvent.list_event_type:
            raise Exception(f"type {type} not in {str(ChatgptEvent.list_event_type)}")
        if event_message == {}:
            raise Exception("Event_message is empty")
        if response_format is not None:
            if event_type != ChatgptEvent.USER_MESSAGE:
                raise Exception("can only format response for user messages")
            if response_format != {"type": "json_object"}:
                list_repsonse_format = [{"type": "json_object"}]
                raise Exception(f"response_format {response_format} is not allowed, must be in {list_repsonse_format}")
        if tool_choice is not None:
            if tool_choice not in ["none", "auto"]:
                raise Exception(f"tool_choice {tool_choice} not in [none, auto]")

        self.event_type = event_type
        self.event_message = event_message
        self.list_tool_offer = list_tool_offer
        self.tool_choice = tool_choice
        self.response_format = response_format

    def is_text_message(self) -> bool:
        return self.event_type in [
            ChatgptEvent.SYSTEM_MESSAGE,
            ChatgptEvent.ASSISTENT_MESSAGE,
            ChatgptEvent.USER_MESSAGE,
        ]

    def __str__(self) -> str:
        if (
            self.event_type == ChatgptEvent.SYSTEM_MESSAGE
            or self.event_type == ChatgptEvent.ASSISTENT_MESSAGE
            or self.event_type == ChatgptEvent.USER_MESSAGE
        ):
            content = self.event_message["content"][0]["text"]  # type: ignore
            if self.event_type == ChatgptEvent.SYSTEM_MESSAGE:
                return f"System message:\n    {content}"
            elif self.event_type == ChatgptEvent.ASSISTENT_MESSAGE:
                return f"Assistent message:\n    {content}"
            else:
                return f"User message:\n    {content}"
        elif self.event_type == ChatgptEvent.TOOL_CALL_REQUEST:
            return f"Tool call request:\n    {json.dumps(self.event_message, indent=4)}"
        elif self.event_type == ChatgptEvent.TOOL_CALL_RESULT:
            return f"Tool call result:\n    {json.dumps(self.event_message, indent=4)}"
        else:
            return f"Event Type: {self.event_type}, Event Content: {self.event_message}"

    def to_dict(self) -> dict:
        dict_event = {
            "type": self.event_type,
            "message": self.event_message,
        }
        if self.list_tool_offer is not None:
            dict_event["list_tool_offer"] = self.list_tool_offer
        if self.tool_choice is not None:
            dict_event["tool_choice"] = self.tool_choice
        if self.response_format is not None:
            dict_event["response_format"] = self.response_format
        return dict_event

    @staticmethod
    def from_dict(event_dict: dict) -> "ChatgptEvent":
        event_type = event_dict["type"]
        event_message = event_dict["message"]
        list_tool_offer = event_dict.get("list_tool_offer")
        tool_choice = event_dict.get("tool_choice")
        response_format = event_dict.get("response_format")
        return ChatgptEvent(
            event_type,
            event_message,
            list_tool_offer=list_tool_offer,
            tool_choice=tool_choice,
            response_format=response_format,
        )


class PromptConfig:
    def __init__(
        self,
        model_id: str,
        list_event: List[ChatgptEvent],
        tool_choice: Optional[str] = None,
    ) -> None:

        if model_id is None:
            raise Exception("model is None")
        if list_event is None:
            raise Exception("list_event is None")
        self.model_id = model_id
        self.list_event = copy(list_event)

    @property
    def last_event(self) -> ChatgptEvent:
        return self.list_event[-1]

    @property
    def tools(self) -> Optional[List[dict]]:
        return self.last_event.list_tool_offer

    @property
    def tool_choice(self) -> Optional[str]:
        return self.last_event.tool_choice

    @property
    def response_format(self) -> Optional[dict]:
        return self.last_event.response_format

    @property
    def tool_calls(self) -> List[dict]:
        if self.last_event.event_type != ChatgptEvent.TOOL_CALL_REQUEST:
            raise Exception("last event is not a tool call request")
        if self.last_event.event_message is None:
            raise Exception("last event message is None")
        return self.last_event.event_message["tool_calls"]

    @property
    def last_message_content(self) -> str:
        return self.messages[-1]["content"]

    @property
    def last_message_text(self) -> str:
        return self.list_text_message[-1]["content"][0]["text"]

    @property
    def list_text_message(self) -> List[dict]:
        list_message = []
        for event in self.list_event:
            if event.is_text_message():
                list_message.append(event.event_message)
        return list_message

    @property
    def messages(self) -> List[dict]:
        return [event.event_message for event in self.list_event]  # type: ignore

    @staticmethod
    def create_message_text(role: str, message_text: str) -> dict:
        message_content = [
            {
                "type": "text",
                "text": message_text,
            },
        ]
        return PromptConfig.create_message(role, message_content)

    @staticmethod
    def create_message(role: str, message_content: List[dict]) -> dict:
        if role not in ["system", "user", "assistant"]:
            raise Exception(f"role {role} not in [system, user, assistant]")
        return {"role": role, "content": message_content}

    def append_system_message(self, message_content: str) -> "PromptConfig":
        event_message = PromptConfig.create_message_text("system", message_content)
        event = ChatgptEvent(ChatgptEvent.SYSTEM_MESSAGE, event_message=event_message)
        prompt_config = PromptConfig(self.model_id, self.list_event)
        prompt_config.list_event.append(event)
        return prompt_config

    def append_user_message(
        self,
        message_text: str,
        image_base64: Optional[str] = None,
        response_format: Optional[dict] = None,
        list_tool_offer: Optional[list[dict]] = None,
        tool_choice: Optional[str] = None,
    ) -> "PromptConfig":
        event_message = PromptConfig.create_message_text("user", message_text)
        if image_base64 is not None:
            if self.model_id != "gpt-4o":
                raise Exception(f"model_id: {self.model_id} not supported for image")
            event_message["content"].append(
                {
                    "type": "image_url",
                    "image_url": {"url": f"data:image/png;base64,{image_base64}"},
                }
            )
        event = ChatgptEvent(
            ChatgptEvent.USER_MESSAGE,
            event_message=event_message,
            response_format=response_format,
            list_tool_offer=list_tool_offer,
            tool_choice=tool_choice,
        )
        prompt_config = PromptConfig(self.model_id, self.list_event)
        prompt_config.list_event.append(event)
        return prompt_config

    def append_assistent_message(
        self, message: ChatCompletionMessage, list_tool_call_result: List[dict] = []
    ) -> "PromptConfig":
        if message.role != "assistant":
            raise Exception("message role is not assistant")
        message_content = message.model_dump()["content"]
        if type(message_content) is str:
            event_message = PromptConfig.create_message_text("assistant", message_content)
        else:
            event_message = PromptConfig.create_message("assistant", message_content)
        event = ChatgptEvent(ChatgptEvent.ASSISTENT_MESSAGE, event_message=event_message)
        prompt_config = PromptConfig(self.model_id, self.list_event)
        prompt_config.list_event.append(event)
        return prompt_config

    def append_assistent_token(self, token: str) -> "PromptConfig":
        if self.list_event[-1].event_type != ChatgptEvent.ASSISTENT_MESSAGE:
            raise Exception("last event is not an assistent message")
        event_message_text = self.list_event[-1].event_message["content"][0]["text"]
        event_message = PromptConfig.create_message("assistant", event_message_text + token)
        event = ChatgptEvent(ChatgptEvent.ASSISTENT_MESSAGE, event_message=event_message)
        prompt_config = PromptConfig(self.model_id, self.list_event[:-1])
        prompt_config.list_event.append(event)
        return prompt_config

    def append_tool_call_request(
        self,
        message: ChatCompletionMessage,
    ) -> "PromptConfig":
        event = ChatgptEvent(
            ChatgptEvent.TOOL_CALL_REQUEST,
            event_message=message.model_dump(),
        )
        prompt_config = PromptConfig(self.model_id, self.list_event)
        prompt_config.list_event.append(event)
        return prompt_config

    def append_tool_call_result(self, list_tool_call_result: List[dict]) -> "PromptConfig":
        list_event: List[ChatgptEvent] = []
        for tool_call_result in list_tool_call_result:
            message = {
                "role": "tool",
                "tool_call_id": tool_call_result["id"],
                "name": tool_call_result["name"],
                "content": tool_call_result["result"],
            }
            list_event.append(ChatgptEvent(ChatgptEvent.TOOL_CALL_RESULT, event_message=message))
        prompt_config = PromptConfig(self.model_id, self.list_event)
        prompt_config.list_event.extend(list_event)
        return prompt_config

    def token_count(self) -> int:
        content = ""
        for message in self.messages:
            content += message["content"]
        if self.model_id == "gpt-4o":
            encoding_name = "cl100k_base"
        elif self.model_id == "gpt-4":
            encoding_name = "cl100k_base"
        elif self.model_id == "gpt-4-turbo-preview":
            encoding_name = "cl100k_base"
        elif self.model_id == "gpt-3.5-turbo":
            encoding_name = "cl100k_base"
        else:
            raise Exception(f"model {self.model_id} not supported")
        encoding = tiktoken.get_encoding(encoding_name)
        num_tokens = len(encoding.encode(content))
        return num_tokens

    def token_count_max(self) -> int:
        if self.model_id == "gpt-4":
            return 128000
        elif self.model_id == "gpt-4-turbo-preview":
            return 128000
        elif self.model_id == "gpt-3.5-turbo":
            return 16385
        else:
            raise Exception(f"model {self.model_id} not supported")

    @staticmethod
    def create(
        model_id: str,
        system_message_text: str,
    ) -> "PromptConfig":
        event_message = PromptConfig.create_message_text("system", system_message_text)
        event = ChatgptEvent("system_message", event_message=event_message)

        return PromptConfig(model_id, [event])

    def to_dict(self) -> dict:
        return {
            "model_id": self.model_id,
            "list_event": [event.to_dict() for event in self.list_event],
        }

    @staticmethod
    def from_dict(prompt_config_dict: dict) -> "PromptConfig":
        model_id = prompt_config_dict["model_id"]
        list_event = [ChatgptEvent.from_dict(event_dict) for event_dict in prompt_config_dict["list_event"]]
        return PromptConfig(model_id, list_event)

    def __str__(self) -> str:
        string_description = ""
        for event in self.list_event:
            string_description += str(event) + "\n"
        return string_description



import inspect
from enum import Enum
from typing import Any, Callable, Dict, List, Literal, Optional, Tuple

import docstring_parser


class ChatgptParameter:
    def __init__(
        self,
        name: str,
        type: str,
        is_required: bool,
        description: Optional[str] = None,
        list_enum_value: Optional[List[str]] = None,
    ) -> None:
        self.name = name
        self.type = type
        self.is_required = is_required

        self.description = description
        self.list_enum_value = list_enum_value


class ChatgptTool:

    def __init__(
        self,
        callable: Callable,
        name: str,
        description: str,
        list_parameter: List[ChatgptParameter],
    ) -> None:
        self.callable = callable
        self.name = name
        self.description = description
        self.list_parameter = list_parameter
        self.list_secret_key = []

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        return self.call(*args, **kwds)

    def call(self, *args, **kwds) -> Any:
        return self.callable(*args, **kwds)

    def to_tool_dict(self) -> dict:
        tool_dict = {
            "type": "function",
            "function": {
                "name": self.name,
                "description": self.description,
                "parameters": {
                    "type": "object",
                    "properties": {},
                    "required": [],
                },
            },
        }
        for parameter in self.list_parameter:
            tool_dict["function"]["parameters"]["properties"][parameter.name] = {
                "type": parameter.type,
            }
            if parameter.description:
                tool_dict["function"]["parameters"]["properties"][parameter.name]["description"] = parameter.description
            if parameter.list_enum_value:
                tool_dict["function"]["parameters"]["properties"][parameter.name]["enum"] = parameter.list_enum_value
            if parameter.is_required:
                tool_dict["function"]["parameters"]["required"].append(parameter.name)
        return tool_dict


def parse_docstring(callable: Callable) -> Tuple[str, Dict[str, str]]:
    docstring_parsed = docstring_parser.parse(inspect.getdoc(callable))  # type: ignore
    if docstring_parsed.short_description is None:
        raise ValueError("The short description in the docstring is required.")
    description = docstring_parsed.short_description

    # TODO check if it matches the signature
    dict_parameter_description = {}
    for (
        param
    ) in docstring_parsed.params:  # TODO there is also something strange that happens if 1 description is missing
        dict_parameter_description[param.arg_name] = param.description
    return description, dict_parameter_description


def get_parameter_type(annotation: type) -> Tuple[str, Optional[List[str]]]:
    if annotation == str:
        return "string", None
    if annotation == int:
        return "number", None
    if annotation == float:
        return "number", None
    if annotation == bool:
        return "boolean", None
    # check in annotation is a literal
    if hasattr(annotation, "__origin__") and annotation.__origin__ == Literal:
        return "string", [str(value) for value in annotation.__args__]
    if issubclass(annotation, Enum):
        enum_members = [member.value for member in annotation]
        return "string", enum_members
    raise ValueError(f"Unsupported parameter type: {annotation}")


def create_chatgpt_tool(callable: Callable) -> "ChatgptTool":
    name = callable.__name__
    signature = inspect.signature(callable)
    if signature.return_annotation == inspect.Signature.empty:
        raise ValueError("The return annotation is required.")
    if signature.return_annotation != str:
        raise ValueError("The return annotation must be a string.")

    description, dict_parameter_description = parse_docstring(callable)
    list_parameter: List[ChatgptParameter] = []
    for parameter in signature.parameters.values():
        if parameter.annotation == inspect.Signature.empty:
            raise ValueError("All parameters must have an annotation.")
        parameter_name = parameter.name
        parameter_type, parameter_list_enum_value = get_parameter_type(parameter.annotation)
        parameter_is_required = parameter.default == inspect.Parameter.empty
        parameter_description = dict_parameter_description.get(parameter_name, None)

        list_parameter.append(
            ChatgptParameter(
                name=parameter_name,
                type=parameter_type,
                is_required=parameter_is_required,
                description=parameter_description,
                list_enum_value=parameter_list_enum_value,
            )
        )
    return ChatgptTool(callable, name, description, list_parameter)



from typing import Literal, Optional

from openai import OpenAI
from srai_core.tools_env import get_string_from_env


class ClientOpenaiAudio:
    def __init__(self, api_key: Optional[str] = None):
        if api_key is None:
            api_key = get_string_from_env("OPENAI_API_KEY")
        self.client_openai = OpenAI(api_key=api_key)

    def get_default_model_id(self) -> str:
        return "whisper-1"

    def get_default_tts_model_id(self) -> str:
        return "tts-1"

    def transcription(
        self,
        path_file_audio: str,
    ) -> dict:
        # TODO add script
        model_id = self.get_default_model_id()
        with open(path_file_audio, "rb") as file:
            transcription = self.client_openai.audio.transcriptions.create(
                model=model_id, file=file, response_format="verbose_json", timestamp_granularities=["word"]
            )
        return transcription.model_dump()

    def text_to_speech_for_file(
        self,
        text: str,
        path_file_audio: str,
        *,
        voice: Literal["alloy", "echo", "fable", "onyx", "nova", "shimmer"] = "alloy",
        response_format: Literal["mp3", "opus", "aac", "flac"] = "mp3",
        model_id: Optional[str] = None,
    ) -> None:
        if model_id is None:
            model_id = self.get_default_tts_model_id()
        speech = self.client_openai.audio.speech.create(
            model=model_id,
            voice=voice,
            response_format=response_format,
            input=text,
        )
        speech.write_to_file(path_file_audio)

    def text_to_speech_for_bytes(
        self,
        text: str,
        *,
        voice: Literal["alloy", "echo", "fable", "onyx", "nova", "shimmer"] = "alloy",
        response_format: Literal["mp3", "opus", "aac", "flac"] = "mp3",
        model_id: Optional[str] = None,
    ) -> bytes:
        if model_id is None:
            model_id = self.get_default_tts_model_id()
        speech = self.client_openai.audio.speech.create(
            model=model_id,
            voice=voice,
            response_format=response_format,
            input=text,
        )
        from io import BytesIO

        bytes_io = BytesIO()
        for chunk in speech.iter_bytes():
            bytes_io.write(chunk)
        bytes_io.seek(0)
        return bytes_io.read()



from typing import List, Optional

from openai import OpenAI
from srai_core.tools_env import get_string_from_env


class ClientOpenaiEmbedding:
    def __init__(self, api_key: Optional[str] = None):
        if api_key is None:
            api_key = get_string_from_env("OPENAI_API_KEY")
        self.client_openai = OpenAI(api_key=api_key)

    def get_default_model_id(self) -> str:
        return "text-embedding-3-small"

    def list_model_id(self) -> list:
        return ["text-embedding-3-small", "text-embedding-3-large"]

    def get_embedding(
        self,
        text: str,
        *,
        model_id: Optional[str] = None,
    ) -> List[float]:
        if model_id is None:
            model_id = self.get_default_model_id()
        embedding = self.client_openai.embeddings.create(input=text, model=model_id)
        return embedding.data[0].embedding



import json
from typing import Dict, List, Optional, Tuple

from openai import OpenAI
from srai_core.tools_env import get_string_from_env

from srai_openai.model.chatgpt_tool import ChatgptTool
from srai_openai.model.prompt_config import ChatgptEvent, PromptConfig


class ClientOpenaiChatgpt:

    def __init__(self, api_key: Optional[str] = None):
        if api_key is None:
            api_key = get_string_from_env("OPENAI_API_KEY")
        self.client_openai = OpenAI(api_key=api_key)

    def get_default_model_id(self) -> str:
        return "gpt-4o"

    def list_model_id(self) -> list:
        model_list = self.client_openai.models.list().data
        return [model.id for model in model_list]

    def prompt_default(
        self,
        system_message_content: str,
        user_message_content: str,
        *,
        model: Optional[str] = None,
        image_base64: Optional[str] = None,
    ) -> str:
        if model is None:
            model = self.get_default_model_id()

        prompt_config_input = PromptConfig.create(model, system_message_content)
        prompt_config_input = prompt_config_input.append_user_message(user_message_content, image_base64)
        return self.prompt_for_prompt_config(prompt_config_input).last_message_text

    def prompt_default_json(
        self,
        system_message_content: str,
        user_message_content: str,
        *,
        model: Optional[str] = None,
        image_base64: Optional[str] = None,
    ) -> dict:

        if model is None:
            model = self.get_default_model_id()

        prompt_config_input = PromptConfig.create(model, system_message_content)
        prompt_config_input = prompt_config_input.append_user_message(
            user_message_content, image_base64, response_format={"type": "json_object"}
        )
        return json.loads(self.prompt_for_prompt_config(prompt_config_input).last_message_text)

    def complete_tool_call_requests(
        self, dict_chatgpt_tool: Dict[str, ChatgptTool], prompt_config_input: PromptConfig
    ) -> PromptConfig:
        if prompt_config_input.list_event[-1].event_type != ChatgptEvent.TOOL_CALL_REQUEST:
            raise Exception("last event is not a tool call request")

        list_tool_call_result = []
        for tool_call_request in prompt_config_input.list_event[-1].event_message["tool_calls"]:
            id = tool_call_request["id"]
            name = tool_call_request["function"]["name"]
            arguments = json.loads(tool_call_request["function"]["arguments"])

            chatgpt_tool = dict_chatgpt_tool[name]

            tool_call_result = {
                "id": id,
                "name": name,
                "result": chatgpt_tool.call(**arguments),
            }
            list_tool_call_result.append(tool_call_result)
        return prompt_config_input.append_tool_call_result(list_tool_call_result)

    def prompt_default_tool(
        self,
        system_message_content: str,
        user_message_content: str,
        list_chatgpt_tool: List[ChatgptTool],
        *,
        image_base64: Optional[str] = None,
        tool_choice: Optional[str] = "auto",
        model: Optional[str] = None,
    ) -> str:
        if model is None:
            model = self.get_default_model_id()

        dict_chatgpt_tool = {}
        list_tool_offer = []
        for chatgpt_tool in list_chatgpt_tool:
            dict_chatgpt_tool[chatgpt_tool.name] = chatgpt_tool
            list_tool_offer.append(chatgpt_tool.to_tool_dict())
        prompt_config_input = PromptConfig.create(model, system_message_content)
        prompt_config_input = prompt_config_input.append_user_message(
            user_message_content,
            image_base64=image_base64,
            list_tool_offer=list_tool_offer,
            tool_choice=tool_choice,
        )

        prompt_config_result = self.prompt_for_prompt_config(prompt_config_input)
        if prompt_config_result.list_event[-1].event_type != ChatgptEvent.TOOL_CALL_REQUEST:
            return prompt_config_result.last_message_text

        prompt_config_input_tools = self.complete_tool_call_requests(dict_chatgpt_tool, prompt_config_result)
        prompt_config_result_tools = self.prompt_for_prompt_config(prompt_config_input_tools)
        return prompt_config_result_tools.last_message_text

    def prompt_for_prompt_config(self, prompt_config_input: PromptConfig) -> PromptConfig:
        completion = self.client_openai.chat.completions.create(
            model=prompt_config_input.model_id,
            messages=prompt_config_input.messages,  # type: ignore
            tools=prompt_config_input.tools,  # type: ignore
            tool_choice=prompt_config_input.tool_choice,  # type: ignore
            response_format=prompt_config_input.response_format,  # type: ignore
        )
        if completion.choices[0].finish_reason == "tool_calls":
            return prompt_config_input.append_tool_call_request(
                completion.choices[0].message,
            )
        else:
            return prompt_config_input.append_assistent_message(
                completion.choices[0].message,
            )

    def options_for_prompt_config(
        self, prompt_config_input: PromptConfig, max_tokens: Optional[int] = 1
    ) -> List[Tuple[float, PromptConfig]]:
        completion = self.client_openai.chat.completions.create(
            model=prompt_config_input.model_id,
            messages=prompt_config_input.messages,  # type: ignore
            logprobs=True,
            top_logprobs=20,
            max_tokens=1,
        )
        model_dump = completion.choices[0].model_dump()
        list_option = []
        for option in model_dump["logprobs"]["content"][0]["top_logprobs"]:
            prompt_config_result = prompt_config_input.append_assistent_message(option["token"])
            list_option.append((option["logprob"], prompt_config_result))
        return list_option

    def prompt_default_image(self, prompt_config_input: PromptConfig) -> str:
        prompt_config_result = self.prompt_for_prompt_config(prompt_config_input)
        return prompt_config_result.last_message_text
